#!/opt/soracom/python/bin/python

# ======================================================
# Project Name    : SORACOM Mosaic CameraApp0 Sample
# File Name       : CameraApp0.py
# Copyright       : 2015-2021 SORACOM, INC.
# ======================================================

"""
NOTE
 - IDEA: 
    - Change the Video Resolution via API
        - http://127.0.0.1:8080/v1/stopCameraCapture
        - http://127.0.0.1:8080/v1/cameraState -X PATCH -H "Content-Type: application/json" -d "{\"width\":640, \"height\":480}"

 - USEFUL API:
    - curl --verbose http://127.0.0.1:8080/v1/deviceState
"""

import os
import time
from time import sleep
from datetime import datetime, timedelta
import json
import re
import logging
from logging import getLogger
import traceback
import cv2

from soracom_device_interface import SoracomDeviceInterface
from soracom_image_util import SoracomImageUtil
from soracom_image_analize_thread_controller import SoracomImageAnalizeThreadController
from soracom_data_upload_thread_controller import SoracomDataUploadThreadController

LOGGING_LEVEL = logging.DEBUG
FORMAT = '%(asctime)s %(levelname)s %(funcName)s %(filename)s %(lineno)d %(message)s'
try:
    root_logger = logging.getLogger()
    root_logger.setLevel(LOGGING_LEVEL)
    root_handler = root_logger.handlers[0]
    root_handler.setFormatter(logging.Formatter(FORMAT))
except IndexError:
    logging.basicConfig(level=LOGGING_LEVEL, format=FORMAT)
logger = getLogger(__name__)

DEBUG_MODE = False

DEVICE_INTERFACE_URI = os.getenv('DEVICE_INTERFACE_URI', None)

ENDPOINT = os.getenv('SORACOM_ENV_ENDPOINT', None)

# image upload interval. too short interval increases upload data volume.
WAIT = int(os.getenv('SORACOM_ENV_WAIT', '60'))
CROP_TOP = int(os.getenv('SORACOM_ENV_CROP_TOP', '10'))
CROP_BOTTOM = int(os.getenv('SORACOM_ENV_CROP_BOTTOM', '10'))
CROP_LEFT = int(os.getenv('SORACOM_ENV_CROP_LEFT', '10'))
CROP_RIGHT = int(os.getenv('SORACOM_ENV_CROP_RIGHT', '10'))
UPLOAD_SIZE_W = int(os.getenv('SORACOM_ENV_UPLOAD_SIZE_W', '640'))
UPLOAD_SIZE_H = int(os.getenv('SORACOM_ENV_UPLOAD_SIZE_H', '480'))
FREE_PARAM = str(os.getenv('SORACOM_ENV_FREE_PARAM', '{}'))

# set DEVICE_INTERFACE_URI, if it is not set.
if DEVICE_INTERFACE_URI is None:
    DEVICE_INTERFACE_URI = 'http://127.0.0.1:8080'

try:
    FREE_PARAM_OBJECT = json.loads(FREE_PARAM)
except Exception:
    FREE_PARAM_OBJECT = {}
    logger.error("free param's json load error")


# choose upload image format, image quality and image decrease factor
IMAGE_FORMAT = FREE_PARAM_OBJECT.get('IForm', 'webp')
IMAGE_QUALITY = FREE_PARAM_OBJECT.get('IQl', 100)
IMAGE_RESIZE = FREE_PARAM_OBJECT.get('IDf', 2)

UPLOAD_TYPE = IMAGE_FORMAT
UPLOAD_QUALITY = IMAGE_QUALITY

# the number of process
PROCESS_NUM = FREE_PARAM_OBJECT.get(
    'MPnum', 1)

# preserve input task order
FIFO = FREE_PARAM_OBJECT.get(
    'FIFO', True)

# process timeout
TIMEOUT = FREE_PARAM_OBJECT.get(
    'TOut', 60)

# data upload interval
UPLOAD_INTERVAL = FREE_PARAM_OBJECT.get(
    'UInv', WAIT)

# image instead of device_interface.getCapture
IMAGE_URIS = FREE_PARAM_OBJECT.get('Images', "")

# adjust contrast and brightness of the camera every AUTO_ADJUST_TIME, if it is set.
AUTO_ADJUST_TIME = 60

# AUTO_ADJUST_TIME can only work for S+ Camera.
if DEVICE_INTERFACE_URI != 'http://127.0.0.1:8080':
    AUTO_ADJUST_TIME = 0

# set fisheye callibration files path
FISHEYE_CALIBRATION_PATH = FREE_PARAM_OBJECT.get('FCalib', None)
if FISHEYE_CALIBRATION_PATH:
    logger.info("FISHEYE_CALIBRATION_PATH: %s", FISHEYE_CALIBRATION_PATH)

# red color correction
RED_CORRECTION = True

# set debug mode
if ENDPOINT is None:
    DEBUG_MODE = True
    ENDPOINT = 'urn:dev:ops:SORACOM-Mosaic-' + os.uname()[1]


# disable multi process when edgetpu is used
# model_name = FREE_PARAM_OBJECT.get(
#     'Model', 'ssd_mobilenet_v2_coco_quant_postprocess.tflite')
# if model_name.count('edgetpu'):
#     PROCESS_NUM = 1

# if PROCESS_NUM == 1:
#     MULTI_PROCESS = False
# else:
#     MULTI_PROCESS = True
MULTI_PROCESS = False

logger.info('DEBUG_MODE = %d', DEBUG_MODE)
logger.info('DEVICE_INTERFACE_URI = %s', DEVICE_INTERFACE_URI)
logger.info('ENDPOINT = %s', ENDPOINT)
logger.info('WAIT = %d', WAIT)
logger.info('CROP_TOP = %d', CROP_TOP)
logger.info('CROP_BOTTOM = %d', CROP_BOTTOM)
logger.info('CROP_LEFT = %d', CROP_LEFT)
logger.info('CROP_RIGHT = %d', CROP_RIGHT)
logger.info('UPLOAD_SIZE_W = %d', UPLOAD_SIZE_W)
logger.info('UPLOAD_SIZE_H = %d', UPLOAD_SIZE_H)
logger.info('FREE_PARAM = %s', FREE_PARAM)
logger.info('PROCESS_NUM = %s', PROCESS_NUM)
logger.info('FIFO = %s', FIFO)
logger.info('TIMEOUT = %s', TIMEOUT)
logger.info('UPLOAD_INTERVAL = %s', UPLOAD_INTERVAL)


IMAGE_UTIL = SoracomImageUtil()
IMAGE_UTIL.set_imageR_Gamma(1.2, 0.8)

MOSAIC_ID = re.sub('^SORACOM-Mosaic-', '', ENDPOINT.split(':')[3])

#######################################
#image settings
IMAGE_UPLOAD_OPTIONS = {}
IMAGE_UPLOAD_OPTIONS['contentType'] = 'image/' + IMAGE_FORMAT
IMAGE_UPLOAD_OPTIONS['imageConvert'] = True          #convert numpy to image
IMAGE_UPLOAD_OPTIONS['quality'] = IMAGE_QUALITY      #image quality
IMAGE_UPLOAD_OPTIONS['topDirName'] = 'beer_can_counter'           #top Harvest Files directory
IMAGE_UPLOAD_OPTIONS['filenamePrefix'] = 'beer_can_counter_' #prefix of Harvest Files
IMAGE_UPLOAD_OPTIONS['filenameSuffix'] = ''           #suffix of Harvest Files file
IMAGE_UPLOAD_OPTIONS['extention'] = IMAGE_FORMAT      #extension for the image

# data settings
DATA_UPLOAD_OPTIONS = {}
DATA_UPLOAD_OPTIONS['upload'] = 'Harvest Data'        #upload destination
#DATA_UPLOAD_OPTIONS['upload'] = 'Harvest Files'
DATA_UPLOAD_OPTIONS['contentType'] = 'application/json'  #content type used for Harvest Files
DATA_UPLOAD_OPTIONS['image_file_path'] = 'image_path' #path for Harvest Files
DATA_UPLOAD_OPTIONS['topDirName'] = 'beer_can_counter'         #top Harvest Files directory
DATA_UPLOAD_OPTIONS['filenamePrefix'] = 'beer_can_counter'   #prefix of Harvest Files
DATA_UPLOAD_OPTIONS['filenameSuffix'] = 'od'           #suffix of Harvest Files
DATA_UPLOAD_OPTIONS['extention'] = 'json'            #extension for the data

UPLOAD_OPTIONS = {}
UPLOAD_OPTIONS['data'] = DATA_UPLOAD_OPTIONS
UPLOAD_OPTIONS['image'] = IMAGE_UPLOAD_OPTIONS
UPLOAD_OPTIONS['debug'] = DEBUG_MODE
#######################################

def Interval():
    logger.info("start")

    # create Soracom Device_interface
    device_interface = SoracomDeviceInterface()
    #device_interface.setCaptureSize(DEVICE_INTERFACE_URI, 1640, 1232, IMAGE_RESIZE)
    #device_interface.setCaptureSize(DEVICE_INTERFACE_URI, 1640, 922, IMAGE_RESIZE)
    # set capture size to 1280x720
    device_interface.setCaptureSize(DEVICE_INTERFACE_URI, 1280, 720, IMAGE_RESIZE)
    #device_interface.setCaptureSize(DEVICE_INTERFACE_URI, 640, 480, IMAGE_RESIZE)
    if FISHEYE_CALIBRATION_PATH:
        device_interface.setUpdateFCalibPath(FISHEYE_CALIBRATION_PATH)

    try:
        if DEBUG_MODE:
            image_analize_thread_controller = SoracomImageAnalizeThreadController(
                1, False, 'soracom_image_analize_object_detection', FIFO, TIMEOUT)
            data_upload_thread_controller = SoracomDataUploadThreadController(
                1, False, 'soracom_data_upload', FIFO, TIMEOUT)
        else:
            image_analize_thread_controller = SoracomImageAnalizeThreadController(
                PROCESS_NUM, MULTI_PROCESS, 'soracom_image_analize_object_detection', FIFO, TIMEOUT)
            data_upload_thread_controller = SoracomDataUploadThreadController(
                PROCESS_NUM, MULTI_PROCESS, 'soracom_data_upload', FIFO, TIMEOUT)
    except Exception as err:
        traceback.print_exc()
        logger.error(
            "fail to create image_analize_thread_controller by %s", err)
        raise

    process_start_time = time.time()
    data_upload_time = 0
    number_of_results = 0
    now_before_time = datetime.now()

    while True:
        try:
            # get an image
            detect_info = None
            if DEBUG_MODE and os.path.exists("./tests/debug.jpg"):
                original_image = cv2.imread("./tests/debug.jpg")
                #original_image = cv2.resize(
                #    original_image, (int(1280/IMAGE_RESIZE), int(720/IMAGE_RESIZE)))
            elif IMAGE_URIS:
                image_index = number_of_results % len(IMAGE_URIS)
                original_image = device_interface.downloadImage(IMAGE_URIS[image_index])
            else:
                original_image = device_interface.getCapture(False)
            if original_image is None:
                sleep(1)
                continue

            # timestamp for the image
            photography_time = time.time()
            analize_image = original_image.copy()
            if RED_CORRECTION:
                analize_image = IMAGE_UTIL.imageR_Gamma(analize_image)

            # send the image to analize function
            if analize_image is not None:
                detect_info = image_analize_thread_controller.putImage(
                    'text', analize_image, photography_time,
                    options={
                        'photography_time': photography_time, 
                        'debug': False,
                    })
                #logger.debug("detect_info: %s", detect_info)

            # adjust contraust and brightness
            if analize_image is not None:
                now = datetime.now()
                if now > now_before_time + timedelta(seconds=AUTO_ADJUST_TIME):
                    IMAGE_UTIL.set_cb(analize_image, device_interface)
                    now_before_time = now

            # create upload data
            if detect_info is not None:
                # extract ocr results
                object_list = detect_info.get('objects', None)
                if object_list is None:
                    continue
                if len(object_list) == 0:
                    continue

                # extract ocr image
                image = detect_info.get('image', None)
                if image is None:
                    continue

                # extract photography_time from the ocr result
                photography_time = time.time()
                if 'options' in detect_info and 'photography_time' in detect_info['options']:
                    photography_time = detect_info['options']['photography_time']

                number_of_results += 1
                logger.debug('number_of_results: %d', number_of_results)

                # wait until to the next upload
                if time.time() - data_upload_time < UPLOAD_INTERVAL:
                    logger.debug(
                        'skip data upload now: %f prv: %f diff: %f',
                        time.time(), data_upload_time, time.time() - data_upload_time)
                    sleep(WAIT)
                    continue
                data_upload_time = time.time()

                ############################################################
                # format upload data
                object_detection = []
                detected_image = []
                for obj in object_list:
                    data = {}
                    data['count'] = obj.get('count', '')
                    data['mosaicId'] = str(MOSAIC_ID)
                    data['ts'] = photography_time
                    object_detection.append(data)
                # add just one image for an object list
                detected_image.append(image)
                result = data_upload_thread_controller.putData(
                    MOSAIC_ID, photography_time, 'default',
                    detected_image, object_detection, time.time(), options=UPLOAD_OPTIONS)
                if result is not None:
                    logger.debug("######## putData result: %s", result)
                sleep(WAIT)
        except Exception as err:
            traceback.print_exc()
            logger.error(err)
            logger.info('sleep: %d', WAIT)
            sleep(WAIT)

def main():
    Interval()


if __name__ == '__main__':
    logger.info('main')
    main()
