# ======================================================
# Project Name    : SORACOM Mosaic CameraApp0 Sample
# File Name       : soracom_image_analize_object_detection.py
# Copyright       : 2015-2020 SORACOM, INC.
# ======================================================

# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#  git clone https://github.com/google-coral/examples-camera.git --depth 1

import glob
import os
import sys
import json
import time
import collections
from datetime import datetime
from logging import getLogger
logger = getLogger(__name__)
import re
import traceback
import platform
import cv2
from PIL import Image
import numpy as np
try:
    import torch
    from torchvision.transforms import Compose

    from yolov5.models.experimental import attempt_load
    from yolov5.utils.datasets import LoadImages
    from yolov5.utils.general import check_img_size, non_max_suppression, scale_coords
    from yolov5.utils.plots import colors, plot_one_box
    from yolov5.utils.torch_utils import select_device
except Exception as e:
    raise Exception("AIライブラリのダウンロードに失敗しました。通信速度を上げて再度インストールをお試しください。")


from soracom_image_util import SoracomImageUtil

COLOR_PURPLE = "\033[35m"
COLOR_RESET = "\033[0m"

IMAGE_UTIL = SoracomImageUtil()

FREE_PARAM = str(os.getenv('SORACOM_ENV_FREE_PARAM', '{}'))

try:
    FREE_PARAM_OBJECT = json.loads(FREE_PARAM)
except Exception:
    FREE_PARAM_OBJECT = {}
    logger.error("free param's json load error")

EDGETPU_SHARED_LIB = {
    'Linux': 'libedgetpu.so.1',
    'Darwin': 'libedgetpu.1.dylib',
    'Windows': 'edgetpu.dll'
}[platform.system()]

Object = collections.namedtuple('Object', ['id', 'score', 'bbox'])

class BBox(collections.namedtuple('BBox', ['xmin', 'ymin', 'xmax', 'ymax'])):
    """Bounding box.
    Represents a rectangle which sides are either vertical or horizontal, parallel
    to the x or y axis.
    """
    __slots__ = ()

class SoracomImageAnalizeObjectDetection():
    def __init__(self):
        """ Run MonoDepthNN to compute depth maps. """
        super().__init__()

        SORACOM_ENV_APP_DATA_ROOT = os.getenv('SORACOM_ENV_APP_DATA_ROOT', '/opt/app/')
        IMAGE_ANALIZE_DIR_NAME = 'soracom_image_analize_model'

        ########################
        weights = os.path.join(SORACOM_ENV_APP_DATA_ROOT, IMAGE_ANALIZE_DIR_NAME,\
            'yolov5s_beer.pt')
         # model.pt path(s)
        imgsz=320 # inference size (pixels)
        device='cpu'# cuda device, i.e. 0 or 0,1,2,3 or cpu
        
        self.half=False # use FP16 half-precision inference
        self.conf_thres=0.2  # confidence threshold
        self.iou_thres=0.45  # NMS IOU threshold
        self.max_det=1000  # maximum detections per image
        self.view_img=False  # show results
        self.save_txt=False  # save results to *.txt
        self.save_conf=False  # save confidences in --save-txt labels
        self.save_crop=False  # save cropped prediction boxes
        self.classes=None  # filter by class: --class 0, or --class 0 2 3
        self.agnostic_nms=False  # class-agnostic NMS
        self.augment=False  # augmented inference
        self.visualize=False  # visualize features
        self.update=False  # update all models
        self.line_thickness=3  # bounding box thickness (pixels)
        self.hide_labels=False  # hide labels
        self.hide_conf=True  # hide confidences
        self.bs = 1  # batch size

        # Initialize
        self.device = select_device(device)
        logger.info(f"{COLOR_PURPLE}device: {self.device}{COLOR_RESET}")

        # Load model
        self.model = attempt_load(weights, map_location=self.device)  # load FP32 model
        self.stride = int(self.model.stride.max())  # model stride
        self.imgsz = check_img_size(imgsz, s=self.stride)  # check image size
        self.names = self.model.module.names if hasattr(self.model, 'module') else self.model.names  # get class names
        if self.half:
            self.model.half()  # to FP16

        ############################
        # # Test Images
        # self.replace_image = True
        # self.has_debug_image = False
        # self.debug_images = []
        # # test_data_path = os.path.join(SORACOM_ENV_APP_DATA_ROOT, IMAGE_ANALIZE_DIR_NAME, 
        # #     'test_peopl_counter_*')
        # for img in glob.glob("./*.jpg"):
        #     im = cv2.imread(img)
        #     self.debug_images.append(im)
        # if len(self.debug_images) > 0:
        #     self.has_debug_image = True

    def saveFile(self, image, filename="object.jpg", note=''):
        _, im_data = cv2.imencode('.jpeg', image)
        fullpath = os.path.join('./', filename)
        try:
            logger.info('image file: %s note: %s', fullpath, note)
            with open(fullpath, mode='wb') as fout:
                fout.write(im_data)
        except FileExistsError as error:
            raise error

    def analizeImage(self, analize_type, image, image_id=-1, resources=None, options={}):
        logger.debug('start analizeImage analize_type: %s image_id: %d resources: %s \
                     options: %s', analize_type, image_id, resources, options)
        debug_mode = options.get('debug', False)

        # #######
        # debug_mode = True
        # if debug_mode:
        #     if self.replace_image and self.has_debug_image:
        #         logger.debug(f"{COLOR_PURPLE}Replacing image{COLOR_RESET}")
        #         image = self.debug_images.pop()
        #         if len(self.debug_images) == 0:
        #             self.has_debug_image = False
        #     else:
        #         logger.debug(f"{COLOR_PURPLE}Could not replace image: {self.replace_image}, {self.has_debug_image}{COLOR_RESET}")
        #         time.sleep(1000000)
        # #######

        photography_time = options.get('photography_time', None)

        results = {}
        start_ms = time.time()
        logger.info("start analizeImage at: %s", start_ms)
        cv2_im = image
        cv2_im_rgb = cv2.cvtColor(cv2_im, cv2.COLOR_BGR2RGB)
        pil_im = Image.fromarray(cv2_im_rgb)
        objs = {'count': 0}

        # ###############################################
        # Dataloader
        # if webcam:
        #     view_img = check_imshow()
        #     cudnn.benchmark = True  # set True to speed up constant image size inference
        #     dataset = LoadStreams(source, img_size=imgsz, stride=stride)
        #     bs = len(dataset)  # batch_size
        # else:
        dataset = LoadImages(cv2_im, img_size=self.imgsz, stride=self.stride)
        bs = 1  # batch_size

        vid_path, vid_writer = [None] * self.bs, [None] * self.bs

        # Run inference
        if self.device.type != 'cpu':
            self.model(torch.zeros(1, 3, self.imgsz, self.imgsz).to(self.device).type_as(next(self.model.parameters())))  # run once
        t0 = time.time()

        for _, img, im0s, _ in dataset:
            img = torch.from_numpy(img).to(self.device)
            img = img.half() if self.half else img.float()  # uint8 to fp16/32
            img /= 255.0  # 0 - 255 to 0.0 - 1.0
            if img.ndimension() == 3:
                img = img.unsqueeze(0)

            # Inference
            pred = self.model(img,
                            augment=self.augment,
                            visualize=False)[0]

            # Apply NMS
            pred = non_max_suppression(pred, self.conf_thres, self.iou_thres, self.classes, self.agnostic_nms, max_det=self.max_det)

            # # Apply Classifier
            # if classify:
            #     pred = apply_classifier(pred, modelc, img, im0s)

            # Process detections
            s, im0 = '', im0s.copy()
            for _, det in enumerate(pred):  # detections per image
                # if webcam:  # batch_size >= 1
                #     p, s, im0, frame = path[i], f'{i}: ', im0s[i].copy(), dataset.count
                # else:
                if len(det):
                    s += '%gx%g ' % img.shape[2:]  # print string

                    # Rescale boxes from img_size to im0 size
                    det[:, :4] = scale_coords(img.shape[2:], det[:, :4], im0.shape).round()

                    # Print results
                    for c in det[:, -1].unique():
                        n = (det[:, -1] == c).sum()  # detections per class
                        s += f"{n} {self.names[int(c)]}{'s' * (n > 1)}, "  # add to string

                    # Write results
                    for *xyxy, conf, cls in reversed(det):
                        c = int(cls)  # integer class
                        class_name = self.names[c]
                        label = None if self.hide_labels else (class_name if self.hide_conf else f'{class_name} {conf:.2f}')
                        logger.debug(f"{COLOR_PURPLE}{label}{COLOR_RESET}")
                        plot_one_box(xyxy, im0, label=label, color=colors(c, True), line_thickness=self.line_thickness)
                        objs['count'] += 1

        # logger.debug(f"{COLOR_PURPLE}{pred}{COLOR_RESET}")

        # ###############################################
        logger.debug(f"{COLOR_PURPLE}{objs}{COLOR_RESET}")
        # Print time (inference + NMS)
        inference_time = (time.time() - start_ms)
        logger.info(f"{COLOR_PURPLE}{inference_time}{COLOR_RESET}")

        if debug_mode:
            self.saveFile(im0, filename=f"{start_ms}.jpg")

        # Print time (inference + NMS)
        inference_time = (time.time() - start_ms)
        logger.info(f"{COLOR_PURPLE}{inference_time}{COLOR_RESET}")

        results['image'] = im0

        if debug_mode:
            # self.saveFile(image)
            pass

        results['objects'] = [objs]
        results['options'] = options
        results['image_id'] = image_id
        results['photography_time'] = photography_time

        logger.info("end analizeImage at: %s elapsed time: %s", time.time(), inference_time)

        return results


SORACOM_IMAGE_ANALIZE_OBJECT_DETECTION = None

def run(analize_type, image, image_id, resources, options):
    global SORACOM_IMAGE_ANALIZE_OBJECT_DETECTION
    if SORACOM_IMAGE_ANALIZE_OBJECT_DETECTION is None:
        SORACOM_IMAGE_ANALIZE_OBJECT_DETECTION = SoracomImageAnalizeObjectDetection()
    result = SORACOM_IMAGE_ANALIZE_OBJECT_DETECTION.analizeImage(
        analize_type, image, image_id, resources, options)
    return result
