# ======================================================
# Project Name    : SORACOM Mosaic CameraApp0 Sample
# File Name       : soracom_movie_upload_thread.py
# Copyright       : 2015-2021 SORACOM, INC.
# ======================================================

from time import sleep
import threading
from multiprocessing import Pool
import importlib

from logging import getLogger

logger = getLogger(__name__)

MOVIE_UPLOAD = None
MOVIE_UPLOAD_MODULE = ''

def UploadMoviePoolRun(upload_data_pool_module,
                      device_id, timestamp, data_type, data, file_path, file_id, resources, options):
    global MOVIE_UPLOAD
    global MOVIE_UPLOAD_MODULE
    result = None
    logger.debug("MOVIE_UPLOAD_MODULE is %s.", MOVIE_UPLOAD_MODULE)
    if upload_data_pool_module is not None and len(upload_data_pool_module) > 0:
        if MOVIE_UPLOAD_MODULE != upload_data_pool_module:
            logger.info("import data upload module name is %s.", upload_data_pool_module)
            MOVIE_UPLOAD = importlib.import_module(upload_data_pool_module)
            if MOVIE_UPLOAD is not None:
                MOVIE_UPLOAD_MODULE = upload_data_pool_module
    if MOVIE_UPLOAD is not None:
        result = MOVIE_UPLOAD.run(device_id, timestamp, data_type, data,
                                  file_path, file_id, resources, options)
    return result

def UploadMoviePool(args):
    result = None
    upload_data_pool_module = args[0]
    device_id = args[1]
    timestamp = args[2]
    data_type = args[3]
    data = args[4]
    file_path = args[5]
    file_id = args[6]
    resources = args[7]
    options = args[8]
    result = UploadMoviePoolRun(upload_data_pool_module,
                               device_id, timestamp, data_type,
                               data, file_path, file_id, resources, options)
    return result

class SoracomMovieUploadThread(threading.Thread):
    def __init__(self, useMultiprocessing=True,
                 upload_data_pool_module='', multiprocessingTimeout=10):
        super().__init__()
        self.upload_data_pool_module = upload_data_pool_module
        if useMultiprocessing:
            self.multiprocessingPool = Pool(processes=1)
        else:
            self.multiprocessingPool = None

        self.locker = threading.Lock()
        self.condition = threading.Condition()
        self.conditionWait = False

        self.device_id = None
        self.timestamp = None
        self.data_type = None
        self.data = None
        self.file_path = None
        self.data_id = None
        self.privacy = False
        self.result = None
        self.working = False
        self.options = None
        self.resources = None
        self.multiprocessingTimeout = multiprocessingTimeout
        # thread demon
        self.daemon = True
        # start thread
        self.start()

    def _startUpload(self, device_id, timestamp, data_type,
                     data, file_path, file_id, resources, options):
        self.device_id = device_id
        self.timestamp = timestamp
        self.data_type = data_type
        self.data = data
        self.file_path = file_path
        self.file_id = file_id
        self.resources = resources
        self.options = options
        self.result = None
        self.working = True
        self.condition.acquire()
        if self.conditionWait:
            self.condition.notify()
        self.condition.release()

    def _getWorking(self):
        return self.working

    def _getResult(self):
        return self.result

    def _resetUpload(self):
        self.device_id = None
        self.timestamp = None
        self.data_type = None
        self.data = None
        self.file_path = None
        self.file_id = None
        self.privacy = False
        self.result = None

    def resetUpload(self):
        self.locker.acquire()
        try:
            self._resetUpload
        except BaseException:
            self.locker.release()
        self.locker.release()

    def getWorking(self):
        result = None
        self.locker.acquire()
        try:
            result = self._getWorking()
        except BaseException:
            self.locker.release()
            return None
        self.locker.release()
        return result

    def getResult(self):
        result = None
        self.locker.acquire()
        try:
            result = self._getResult()
        except BaseException:
            self.locker.release()
            return None
        self.locker.release()
        return result

    def startUpload(self, device_id, timestamp, data_type,
                    data, file_path, file_id=-1, resources=None, options={}):
        self.locker.acquire()
        try:
            self._startUpload(device_id, timestamp, data_type,
                              data, file_path, file_id, resources, options)
        except BaseException:
            self.locker.release()
            return
        self.locker.release()
        return

    def _uploadMovieThread(self):
        #Lock
        logger.debug('_uploadMovieThread start')
        self.locker.acquire()
        device_id = self.device_id
        timestamp = self.timestamp
        data_type = self.data_type
        data = self.data
        file_path = self.file_path
        file_id = self.file_id
        resources = self.resources
        options = self.options
        self.locker.release()
        logger.debug("uploadMovie start %s", file_id)
        if self.multiprocessingPool is not None:
            res = self.multiprocessingPool.apply_async(UploadMoviePool, \
                  [(self.upload_data_pool_module, device_id, timestamp,
                    data_type, data, file_path, file_id, resources, options)])
            try:
                # Wait timeout seconds for func to complete.
                result = res.get(self.multiprocessingTimeout)
            except TimeoutError:
                res.revoke()
                logger.error("id %d is aborted due to timeout", file_id)
                return None
        else:
            result = UploadMoviePoolRun(self.upload_data_pool_module, \
                    device_id, timestamp, data_type, data, file_path, file_id, resources, options)
        logger.debug("uploadMovie end %s", file_id)
        self.locker.acquire()
        self.result = result
        self.locker.release()
        logger.debug('_uploadMovieThread end')
        return True

    def _threadMain(self):
        while True:
            try:
                self.condition.acquire()
                self.conditionWait = True
                self.condition.wait(None)
                self.conditionWait = False
                self.condition.release()
                logger.info('uploadMovie')
                try:
                    self._uploadMovieThread()
                except Exception as err:
                    import traceback
                    traceback.print_exc()
                    logger.error(err)
                    sleep(5)
                self.locker.acquire()
                self.working = False
                self.locker.release()
            except Exception as err:
                import traceback
                traceback.print_exc()
                logger.error(err)
                sleep(5)

    def run(self):
        self._threadMain()
